/*
 * $Id: mkind.h,v 1.14 1992/01/03 13:21:40 schrod Exp $
 *----------------------------------------------------------------------
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 *
 * [history at end]
 */

/***********************************************************************

                           INSTALLATION NOTES
                              <21 Jun 90>

   At compile-time (or here, if compile-time definition is not
   available), set non-zero ONE OS_xxxx operating system flag, and if
   more than one compiler is available, ONE compiler flag:
   (The list is sorted alphabetically!)

    Operating Systems        Compiler(s)
    =================        ===========
    OS_ATARI
    OS_BSD
    OS_BS2000                CCD_2000
    OS_MVSXA                 IBM_C370
    OS_PCDOS                 IBM_PC_MICROSOFT or IBM_PC_TURBO
    OS_SYSV
    OS_TOPS20                KCC_20 or PCC_20
    OS_VAXVMS
    OS_XENIX
    =================        ===========

   If no OS_xxxx symbol is defined, OS_BSD is assumed.

   If draft ANSI C prototypes are supported, define the symbol
   ANSI_PROTOTYPES in the appropriate OS_xxxx section below, and insert
   #include's for the standard system files which define library
   prototypes.  Unfortunately, these are scattered about in several
   files in draft ANSI C; the IBM_PC_MICROSOFT section below is an
   example which should be close for other implementations.

   All function declarations in MakeIndex are contained at the end of
   this file.  If ANSI_PROTOTYPES is not selected, then all the standard
   library functions must be declared explicitly.

   If the host system restricts external names to 6 characters, set
   SHORTNAMES non-zero in the appropriate OS_xxxx section below.

   Installing MakeIndex under BS2000 requires at least Version 8.5
   supporting Multiple Public Volume Sets (MPVS). Define WORK in this
   file to your CATID (default :w:) for temporary files (.ilg, .ind),
   in case of no MPVS support define WORK to NIL.

*/

#ifndef _MKIND_H_
#define _MKIND_H_

/*======================================================================
 * 
 * 	CONFIGURATION SECTION
 * 
 * Change here to adept for your system:
 */

/*
 * This is the default search path for MakeIndex style files on your system.
 * Below is the UNIX version.
 */
#define DEF_STYLE_PATH	".:/usr/local/lib/makeindx"

#ifdef KPATHSEA
#include <kpathsea/config.h>
#include <kpathsea/c-ctype.h>
#endif
/**********************************************************************/
/*
 * Establish needed operating symbols (defaulting to OS_BSD if none
 * specified at compile time). If you add one, add it to the check
 * list at the end of this section, too.
 *
 * [sorted like list above]
 */

#ifndef    OS_ATARI
#define OS_ATARI 0
#endif

#ifndef    OS_BSD
#define OS_BSD 0
#endif

#ifndef    OS_BS2000
#define OS_BS2000 0
#endif

#ifndef    CCD_2000
#define CCD_2000 0
#endif

#ifndef    OS_MVSXA
#define OS_MVSXA 0
#endif

#ifndef    IBM_C370
#define IBM_C370 0
#endif

#ifndef    OS_PCDOS
#define OS_PCDOS 0
#endif

#ifndef    IBM_PC_LATTICE	/* not usable now: see LOG */
#define IBM_PC_LATTICE 0
#endif

#ifndef    IBM_PC_MICROSOFT
#define IBM_PC_MICROSOFT 0
#endif

#ifndef    IBM_PC_TURBO
#define IBM_PC_TURBO 0
#endif

#ifndef    OS_SYSV
#define OS_SYSV 0
#endif

#ifndef    OS_TOPS20
#define OS_TOPS20 0
#endif

#ifndef    KCC_20
#define KCC_20 0
#endif

#ifndef    PCC_20
#define PCC_20 0
#endif

#ifndef    OS_VAXVMS
#define OS_VAXVMS 0
#endif

#ifndef    OS_WIN32
#define OS_WIN32 0
#endif

#ifndef    OS_XENIX
#define OS_XENIX 0
#endif

				/* check list of operating systems */

#if	(OS_ATARI | OS_BSD | OS_BS2000 | OS_MVSXA | OS_PCDOS | OS_SYSV | \
	OS_TOPS20 | OS_VAXVMS | OS_WIN32 | OS_XENIX)
#else
#undef OS_BSD
#define OS_BSD 1                      /* BSD is default operating system */
#endif


/**********************************************************************
 * Establish needed compiler constants
 * [sorted by operating systems]
 */

#if    OS_BS2000
#undef CCD_2000
#define CCD_2000 1                     /* CCD_2000 is default for OS_BS2000 */
#endif

#if    OS_MVSXA
#undef IBM_C370
#define IBM_C370 1                     /* IBM_C370 is default for OS_MVSXA */
#endif

#if    OS_TOPS20
#if    (KCC_20 | PCC_20)
#else
#undef PCC_20
#define PCC_20 1                       /* PCC-20 is default for Tops-20 */
#endif                                 /* KCC_20 | PCC_20 */
#endif                                 /* OS_TOPS20 */


/**********************************************************************
 * Establish compiler characteristics
 */

#if    (OS_TOPS20 | CCD_2000 | OS_MVSXA)
#define SHORTNAMES 1
#else
#define SHORTNAMES 0
#endif

#define SIZE_T	unsigned
#define ANSI_PROTOTYPES 0

/**********************************************************************/

#include    <stdio.h>
#include    <ctype.h>

#if    (CCD_2000 | OS_MVSXA | OS_SYSV)
#include    <string.h>
#endif				/* CCD_2000 | OS_MVSXA | OS_SYSV */

#if    IBM_PC_MICROSOFT
#include    <io.h>                     /* for function declarations */
#endif                                 /* IBM_PC_MICROSOFT */

#if    IBM_PC_TURBO
#include    <io.h>                     /* for function declarations */
#endif                                 /* IBM_PC_TURBO */

#if OS_WIN32
#include <io.h>
#include    <string.h>                 /* for function declarations */
#undef ANSI_PROTOTYPES
#define ANSI_PROTOTYPES    1           /* so we get full argument
                                          type checking */
#define __STDC__ 1
#undef SIZE_T
#define SIZE_T	size_t
#endif /* OS_WIN32 */

#ifdef __STDC__
#include    <stddef.h>                 /* for function declarations */
#ifndef __GNUC__
#include    <stdlib.h>                 /* for function declarations */
#endif
#include    <string.h>                 /* for function declarations */
#undef ANSI_PROTOTYPES
#define ANSI_PROTOTYPES    1           /* so we get full argument
                                          type checking */
#undef SIZE_T
#define SIZE_T	size_t
#endif					/* __STDC__ */

#if    (OS_ATARI | OS_BSD | PCC_20 | OS_XENIX)
#define strchr  index                  /* convert ANSI form to old K&R form  */
#define strrchr rindex                 /* convert ANSI form to old K&R form  */
#endif

#if    IBM_C370
#define access(fn, R_OK) 0             /* function not available */
#endif

#if    PCC_20
#define R_OK 0                         /* PCC-20 access(file,mode) */
#endif                                 /* only understands mode=0  */

#if    SHORTNAMES
/*
 * Provide alternate external names which are unique in the first SIX
 * characters as required for portability (and Oct. 1986 draft ANSI C)
 */
#define check_all          chk_all
#define check_idx          chk_idx
#define check_mixsym       chk_mix
#define compare_one        cmp_one
#define compare_page       cmp_page
#define compare_string     cmp_string
#define delim_n            dlm_n
#define delim_p            dlm_p
#define delim_r            dlm_r
#define delim_t            dlm_t
#define encap_i            ecp_i
#define encap_p            ecp_p
#define encap_range        ecp_range
#define encap_s            ecp_s
#define group_skip         grp_skip
#define group_type         grp_type
#define idx_aclose         idxaclose
#define idx_actual         idxactual
#define idx_keyword        idx_kwd
#define indent_length      ind_length
#define indent_space       ind_space
#define headings_flag      hd_flag
#define heading_pre        hd_pre
#define heading_suf        hd_suf
#define process_idx        prc_idx
#define process_precedence prc_pre
#define range_lc           rng_lc
#define range_ptr          rng_ptr
#define scan_alpha_lower   scnalw
#define scan_alpha_upper   scnaup
#define scan_arabic        scnarabic
#define scan_arg1          scna1
#define scan_arg2          scna2
#define scan_char          scnchr
#define scan_field         scnfld
#define scan_idx           scnidx
#define scan_key           scnkey
#define scan_no            scnno
#define scan_roman_lower   scnrlw
#define scan_roman_upper   scnrup
#define scan_spec          scnspc
#define scan_string        scnstr
#define scan_style         scnsty
#define setpagelen         spg_len
#define setpage_close      spg_close
#define setpage_open       spg_open
#endif                                 /* SHORTNAMES */

#if    OS_VAXVMS
#define EXIT(code) exit( ((code) == 0) ? 1 : (1 << 28) + 2 )
#else                                  /* NOT OS_VAXVMS */
#define EXIT exit
#endif                                 /* OS_VAXVMS */

#ifndef    R_OK
#define R_OK 4                         /* only symbol from sys/file.h */
#endif

#undef TRUE
#define TRUE 1

#undef FALSE
#define FALSE 0

#undef NUL
#define NUL '\0'

#undef NIL
#define NIL ""

#define TAB '\t'
#define LFD '\n'
#define SPC ' '
#define LSQ '['
#define RSQ ']'
#define BSH '\\'

#if    OS_ATARI		/* (-js) FIXME: what Atari compiler uses these? */
#define ENV_SEPAR ':'	/* should be like in OS_PCDOS !?		*/
#define DIR_DELIM '/'	/* but this was the way it was compiled in 2.9!	*/
#endif                                 /* OS_ATARI */

#if    OS_BSD
#define ENV_SEPAR ':'
#define DIR_DELIM '/'
#endif                                 /* OS_BSD */

#if    OS_BS2000
#define ENV_SEPAR '#'
#define DIR_DELIM '$'
#endif                                 /* OS_BS2000 */

#if    OS_MVSXA
#define ENV_SEPAR '.'
#define DIR_DELIM '.'
#endif                                 /* OS_MVSXA */

#if    OS_PCDOS
#define ENV_SEPAR ';'
#define DIR_DELIM '\\'
#endif                                 /* OS_PCDOS */

#if    OS_SYSV
#define ENV_SEPAR ':'
#define DIR_DELIM '/'
#endif                                 /* OS_SYSV */

#if    OS_TOPS20
#define ENV_SEPAR ','
#define DIR_DELIM ':'
#endif                                 /* OS_TOPS20 */

#if    OS_VAXVMS
#define ENV_SEPAR ','                  /* In fact these aren't used */
#define DIR_DELIM ':'
#endif                                 /* OS_VAXVMS */

#if    OS_WIN32
#define ENV_SEPAR ';'                  /* In fact these aren't used */
#define DIR_DELIM '/'
#endif                                 /* OS_VAXVMS */

#if    OS_XENIX
#define ENV_SEPAR ':'
#define DIR_DELIM '/'
#endif                                 /* OS_XENIX */

#define SW_PREFIX  '-'
#define EXT_DELIM  '.'
#define ROMAN_SIGN '*'
#define EVEN       "even"
#define ODD        "odd"
#define ANY        "any"

#define GET_CHAR getc

#ifndef KPATHSEA
#if    (CCD_2000 | OS_MVSXA)
#define TOASCII(i) (char)((i) + '0')
#else
#define TOASCII(i) (char)((i) + 48)
#endif                                 /* (CCD_2000 | OS_MVSXA) */

#define TOLOWER(C) (char)((isupper(C)) ? tolower(C) : (C))
#define TOUPPER(C) (char)((isupper(C)) ? (C) : toupper(C))

#define STREQ(A, B)  (strcmp(A, B) == 0)
#endif

#define STRNEQ(A, B) (strcmp(A, B) != 0)

#define MESSAGE(F, S) { \
    if (verbose) \
        fprintf(stderr, F, S); \
    fprintf(ilg_fp, F, S); \
}

#ifdef FATAL
#undef FATAL
#define FATAL(F, S) { \
    fprintf(stderr, F, S); \
    fprintf(stderr, USAGE, pgm_fn); \
    EXIT(1); \
}
#endif

#ifdef FATAL2
#undef FATAL2
#define FATAL2(F, D1, D2) { \
    fprintf(stderr, F, D1, D2); \
    fprintf(stderr, USAGE, pgm_fn); \
    EXIT(1); \
}
#endif

#if    OS_PCDOS                        /* open file in binary mode */
#define OPEN_IN_B(FP) fopen(FP, "rb")
#endif

#define OPEN_IN(FP)   fopen(FP, "r")
#define OPEN_OUT(FP)  fopen(FP, "w")
#define CLOSE(FP)     fclose(FP)

#if    (CCD_2000 | OS_MVSXA)
#define ASCII 1
#define ISDIGIT(C)  isdigit(C)
#define ISSYMBOL(C) (! (isdigit(C) || isalpha(C) || iscntrl(C) || C == SPC))
#else
#if WIN32
#ifndef KPATHSEA
#define ISDIGIT(C)  isdigit(C)
#endif
#define ISSYMBOL(C) (! (isdigit(C) || isalpha(C) || iscntrl(C) || C == SPC))
#else
#define ISDIGIT(C)  ('0' <= C && C <= '9')
#define ISSYMBOL(C) (('!' <= C && C <= '@') || \
                     ('[' <= C && C <= '`') || \
                     ('{' <= C && C <= '~'))
#endif
#endif                                 /* (CCD_2000 | OS_MVSXA) */

#if    OS_PCDOS                        /*  512K <= DOS main memory <= 640K */
#define LONG_MAX      144
#define STRING_MAX    64
#else
#ifdef LONG_MAX
#undef LONG_MAX
#define LONG_MAX      1024
#endif
#define STRING_MAX    128
#endif                                 /* OS_PCDOS */

#define ARGUMENT_MAX  1024
#define LINE_MAX      72
#define ARABIC_MAX    5
#define ROMAN_MAX     16
#define PAGETYPE_MAX  5
#define PAGEFIELD_MAX 10
#define FIELD_MAX     3
#define NUMBER_MAX    16

#define VERSION       "version 3.0.8 [31 Dec 91]"
#define PUT_VERSION   MESSAGE("This is MakeIndex, %s.\n", VERSION)

#define USAGE \
   "Usage: %s [-ilqrcg] [-s sty] [-o ind] [-t log] [-p num] [idx0 idx1 ...]\n"

#define STYLE_PATH "INDEXSTYLE"
#define INDEX_IDX  ".idx"
#define INDEX_ILG  ".ilg"
#define INDEX_IND  ".ind"
#if    OS_VAXVMS
#define INDEX_LOG  ".lis"
#else
#define INDEX_LOG  ".log"
#endif

#if    OS_BS2000                       /* Define WORK disc */
#define WORK ":w:"
#define CATID ':'
#endif                                 /* OS_BS2000 */

#define EMPTY     -9999
#define ROML      0
#define ROMU      1
#define ARAB      2
#define ALPL      3
#define ALPU      4
#define DUPLICATE 9999

#define USORTGROUP	256	       /* first user defined group id */

typedef struct KFIELD {
    char    *sf[FIELD_MAX];            /* sort key */
    char    *mf[FIELD_MAX];	       /* merge key */
    char    *af[FIELD_MAX];            /* actual key */
    short   group;		       /* sort group */
    char    lpg[NUMBER_MAX];           /* literal page */
    short   npg[PAGEFIELD_MAX];        /* page field array */
    short   count;                     /* page field count */
    short   type;                      /* page number type */
    char    encap[STRING_MAX];         /* encapsulator */
    char   *fn;                        /* input filename */
    int     lc;                        /* line number */
}       FIELD, *FIELD_PTR;

typedef struct KNODE {
    FIELD   data;
    struct KNODE *next;
}       NODE, *NODE_PTR;

extern int compress_blanks;
extern int init_page;
extern int merge_page;
extern int even_odd;
extern int verbose;

extern char *idx_keyword;
extern char idx_aopen;
extern char idx_aclose;
extern char idx_level;
extern char idx_ropen;
extern char idx_rclose;
extern char idx_quote;
extern char idx_actual;
extern char idx_encap;
extern char idx_escape;

extern char *page_comp;
extern int page_offset[PAGETYPE_MAX];

extern char *preamble;
extern char *postamble;
extern char *setpage_open;
extern char *setpage_close;
extern char *group_skip;
extern int headings_flag;
extern char *heading_pre;
extern char *heading_suf;
extern int prelen;
extern int postlen;
extern int skiplen;
extern int headprelen;
extern int headsuflen;
extern int setpagelen;

extern char *item_r[FIELD_MAX];
extern char *item_u[FIELD_MAX];
extern char *item_x[FIELD_MAX];
extern int ilen_r[FIELD_MAX];
extern int ilen_u[FIELD_MAX];
extern int ilen_x[FIELD_MAX];

extern char *delim_p[FIELD_MAX];
extern char *delim_n;
extern char *delim_r;
extern char *delim_t;

extern char *encap_p;
extern char *encap_i;
extern char *encap_s;

extern int linemax;
extern char *indent_space;
extern int indent_length;

extern FILE *idx_fp;
extern FILE *ind_fp;
extern FILE *ilg_fp;

#if    OS_BS2000
extern FILE *widx_fp;
extern char widx_fn[STRING_MAX + 5];
#endif                                 /* OS_BS2000 */

extern char *idx_fn;
extern char *pgm_fn;
extern char *ind_fn;
extern char *sty_fn;
extern char *ilg_fn;

#ifndef MKIND_C
extern char ind[];
extern char ilg[];
extern char pageno[];

#ifdef DEBUG
extern long totmem;
#endif /* DEBUG */

#endif

extern FIELD_PTR *idx_key;
extern NODE_PTR head;
extern NODE_PTR tail;

extern int idx_dot;
extern int idx_tt;
extern int idx_gt;
extern int idx_et;
extern int idx_dc;

#define DOT     "."
#define DOT_MAX 1000
#define CMP_MAX 1500

#define IDX_DOT(MAX) { \
    idx_dot = TRUE; \
    if (idx_dc++ == 0) { \
        if (verbose) \
            fprintf(stderr, DOT); \
        fprintf(ilg_fp, DOT); \
    } \
    if (idx_dc == MAX) \
        idx_dc = 0; \
}

#define ALL_DONE { \
    if (fn_no > 0) { \
        if (verbose) \
fprintf(stderr, \
        "Overall %d files read (%d entries accepted, %d rejected).\n", \
        fn_no+1, idx_gt, idx_et); \
fprintf(ilg_fp, \
        "Overall %d files read (%d entries accepted, %d rejected).\n", \
        fn_no+1, idx_gt, idx_et); \
    } \
}

#define DONE(A, B, C, D) { \
    if (verbose) \
        fprintf(stderr, "\ndone (%d %s, %d %s).\n", (A), B, C, D); \
    fprintf(ilg_fp, "\ndone (%d %s, %d %s).\n", (A), B, C, D); \
}

#if    ANSI_PROTOTYPES
#define ARGS(arg_list)	arg_list
#define VOIDP		void*
#else
#define ARGS(arg_list)	()
#define const
#define VOIDP		char*
#endif

extern void gen_ind ARGS((void));
extern int main ARGS((int argc, char **argv));
extern void qqsort ARGS((char *base, int n, int size, 
			 int (*compar) ARGS((void*, void*))));
extern void scan_idx ARGS((void));
extern void scan_sty ARGS((char *filename));
extern void set_stylepath ARGS((void));
extern void sort_idx ARGS((void));
extern int strtoint ARGS((char *str));

#include "ordrules.h"

/* Miscellaneous standard library routines */

#ifndef WIN32
VOIDP	calloc ARGS((SIZE_T nitems,SIZE_T size));

#if    CCD_2000                        /* function not available */
#define getenv(P) 0
#else
char   *getenv ARGS((const char *name));
#endif

char   *strchr ARGS((const char *s,int c));
VOIDP   allocate ARGS((SIZE_T size));
VOIDP	malloc ARGS((SIZE_T size));
char   *strrchr ARGS((const char *s,int c));
#endif /* WIN32 */

/* New routines for version 3 */

extern VOIDP fmalloc ARGS((SIZE_T size));



/*======================================================================
 *
 * $Log: mkind.h,v $
 * Revision 1.14  1992/01/03  13:21:40  schrod
 * Declared prototypes.
 *
 * Revision 1.13  1991/12/18  16:22:30  schrod
 * Enhanced handling of style files:
 *    Style file must always be there, a default (mkind.ist) is supplied.
 *    Default extension ".ist".
 *    Default search path to be configured in mkind.h. (Predefined is
 * ".:/usr/local/lib/makeindx", suitable on UNIX systems.)
 *
 * Now issue only short messages while reading style files. In a later
 * version a user should be able to turn long messages on again.
 *
 * Revision 1.12  1991/12/16  14:55:12  schrod
 * gcc does not have stdlib.h.
 * Introduces default style path, set_stylepath() now without params.
 *
 * Revision 1.11  1991/12/10  20:30:39  schrod
 * MakeIndex, version 3.0.6
 * == include in style files ==
 * == group titles may be different for lower/upper-case ==
 *
 * Revision 1.10  1991/12/10  20:26:51  schrod
 * Completely rewritten the scanner to implement a C-like include feature.
 *
 * Revision 1.9  1991/12/10  20:09:54  schrod
 * Reject-rules (*merge_rule and *sort_rule) implemented.
 *
 * Revision 1.8  1991/12/10  19:59:48  schrod
 * Fixed bug in ordrules. Inserting a new group rule at the beginning of
 * a rule list works not with groups, that have the same prefix:
 *
 * 	group_rule "ae" 1
 * 	group_rule "a"  2
 *
 * would put 'aexxx' in group 2.
 *
 * Revision 1.7  1991/12/10  19:49:54  schrod
 * Optimized group-rule handling.
 *
 * Revision 1.6  1991/12/10  19:43:32  schrod
 * Optimized module 'ordrules':
 * - all three types of rules (chr,str,regexp) now allowed for both sort
 *   and merge keys
 *
 * Revision 1.5  1991/12/10  19:20:27  schrod
 * Changed the group merging algorithm to create the group-id of an index key
 * using user defined rules (specified in the style file with commands:
 * `sort_group' and `group_head') instead of the sortkey's first character.
 *
 * Revision 1.4  1991/12/10  17:19:09  schrod
 * New version with flexible index sorting and merging. Style files now
 * can contain rules to describe, how to generate a sort or a merge key
 * from a given index entry.
 *
 * Revision 1.3  1991/12/10  16:53:28  schrod
 *  Changed the representation of an index entry in the KFIELD structure
 * (`af' and `sf') to dynamically allocated strings.
 *
 * Revision 1.2  1991/12/05  20:08:22  schrod
 * MakeIndex, version 2.9.1.2 [by Joachim Schrod]
 * == base version, from which development of International MakeIndex starts ==
 *
 */

#endif
