/*
 * $Id: scanst.c,v 1.9 1992/01/03 13:22:34 schrod Exp $
 *----------------------------------------------------------------------
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 *
 * [history at end]
 */

#include    "mkind.h"
#include    "scanst.h"

#ifdef KPATHSEA
#include <kpathsea/tex-file.h>
#include <kpathsea/lib.h>
#endif

char *stylepath;

char *idx_keyword =	IDX_KEYWORD;
char  idx_aopen =	IDX_AOPEN;
char  idx_aclose =	IDX_ACLOSE;
char  idx_level =	IDX_LEVEL;
char  idx_ropen =	IDX_ROPEN;
char  idx_rclose =	IDX_RCLOSE;
char  idx_quote =	IDX_QUOTE;
char  idx_actual =	IDX_ACTUAL;
char  idx_encap =	IDX_ENCAP;
char  idx_escape =	IDX_ESCAPE;

char *preamble = 	PREAMBLE_DEF;
char *postamble = 	POSTAMBLE_DEF;
int   prelen =		PREAMBLE_LEN;
int   postlen =		POSTAMBLE_LEN;

char *setpage_open =	SETPAGEOPEN_DEF;
char *setpage_close =	SETPAGECLOSE_DEF;
int   setpagelen =	SETPAGE_LEN;

char *group_skip =	GROUPSKIP_DEF;
int   skiplen =		GROUPSKIP_LEN;

int   headings_flag =	HEADINGSFLAG_DEF;
char *heading_pre =	HEADINGPRE_DEF;
char *heading_suf =	HEADINGSUF_DEF;
int   headprelen =	HEADINGPRE_LEN;
int   headsuflen =	HEADINGSUF_LEN;

char *item_r[FIELD_MAX] = {ITEM0_DEF, ITEM1_DEF, ITEM2_DEF};
char *item_u[FIELD_MAX] = {"", ITEM1_DEF, ITEM2_DEF};
char *item_x[FIELD_MAX] = {"", ITEM1_DEF, ITEM2_DEF};

int   ilen_r[FIELD_MAX] = {ITEM_LEN, ITEM_LEN, ITEM_LEN};
int   ilen_u[FIELD_MAX] = {0, ITEM_LEN, ITEM_LEN};
int   ilen_x[FIELD_MAX] = {0, ITEM_LEN, ITEM_LEN};

char *delim_p[FIELD_MAX] = {DELIM_DEF, DELIM_DEF, DELIM_DEF};
char *delim_n = DELIM_DEF; /* page number separator */
char *delim_r = DELIMR_DEF;/* page range designator */
char *delim_t = DELIMT_DEF;/* page list terminating delimiter */

char *encap_p = ENCAP0_DEF;/* encapsulator prefix */
char *encap_i = ENCAP1_DEF;/* encapsulator infix */
char *encap_s = ENCAP2_DEF;/* encapsulator postfix */

int   linemax = LINE_MAX;

int   indent_length = INDENTLEN_DEF;
char *indent_space = INDENTSPC_DEF;

char *page_comp = COMPOSITOR_DEF;
int   page_offset[PAGETYPE_MAX] = {
	0,
	ROMAN_LOWER_OFFSET,
	ROMAN_LOWER_OFFSET + ARABIC_OFFSET,
	ROMAN_LOWER_OFFSET + ARABIC_OFFSET + ALPHA_LOWER_OFFSET,
	ROMAN_LOWER_OFFSET + ARABIC_OFFSET + ALPHA_LOWER_OFFSET + ROMAN_UPPER_OFFSET
};

static int sty_lc = 0;                 /* line count */
static int sty_tc = 0;                 /* total count */
static int sty_ec = 0;                 /* error count */

static int put_dot;

typedef union {
   char s[LONG_MAX];
   char c;
   int d;
} arg_type;

#ifndef WIN32
static FILE *find_style_file ARGS((char *fname));
#else
#define find_style_file(f) kpse_open_file(f, kpse_ist_format)
#endif
static void  scan_file ARGS((FILE *file));
static int   scan_args ARGS((FILE *file, char *format, arg_type *arg));
static int   scan_spec ARGS((FILE *file, char *spec));
static int   scan_char ARGS((FILE *file, char *c));
static int   scan_string ARGS((FILE *file, int unquote, char *str));
static int   scan_int ARGS((FILE *file, int *n));
static int   scan_hexchar ARGS((FILE *file, char *c ));
static void  skip_whitespace ARGS((FILE *file));
static void  skip_line ARGS((FILE *file));
static int   process_precedence ARGS((char *prec));
static int   count_lfd ARGS((char *str));

void
set_stylepath()
{
  char *path;
  register char *r;

  if ((path = getenv(STYLE_PATH)) == (char *) NULL) path = DEF_STYLE_PATH;
  
  stylepath = (char *) fmalloc(strlen(path) + 2);
  strcpy(stylepath,path);

  for (r = stylepath ; *r != '\0' ; r++)
    if (*r == ENV_SEPAR) *r = 0;
    
  *++r = -1;
}

void
scan_sty( filename )
char *filename;
{
   FILE *style_file;

   if ((style_file = find_style_file(filename)) == (FILE *) NULL) {
     STY_ERROR("Style file '%s' not found.",filename);
     return;
   }

   MESSAGE("Scanning style file %s...\n",filename);
   scan_file(style_file);
/* DONE(sty_tc - sty_ec, "attributes redefined", sty_ec, "ignored"); */
   CLOSE(style_file);
}

#ifndef KPATHSEA
static
FILE *find_style_file( fname )
char *fname;
{
   FILE *fd = (FILE *) NULL;
   char *path = stylepath;
   char fullname[LONG_MAX];
   register char *s;
   
   while (*path != -1 && fd == (FILE *) NULL)
   {
      strcpy(fullname,path);
      s = fullname + strlen(fullname);
      *s++ = DIR_DELIM;
      strcpy(s,fname);
      if ( (fd=OPEN_IN(fullname)) == NULL ) {	/* file not found? */
         strcat(s,".ist");			/* append default extension */
	 fd = OPEN_IN(fullname);
      }
      path += strlen(path) + 1;
   }

   return fd;
}
#endif

#define next_char(f)	\
  (actual_char = GET_CHAR(f), (actual_char == LFD ? sty_lc++ : 0), actual_char)

static int actual_char;

static void
scan_file( file )
FILE *file;
{
   static char spec[STRING_MAX];
   static arg_type arg[SPEC_ARG_MAX];
   struct hashentry *entry;

   skip_whitespace(file);
   
   for ( ; actual_char != EOF ; skip_whitespace(file) ) {

/*    STY_DOT; */
      sty_tc++;

      if (!scan_spec(file,spec)) {
      	 skip_line(file);
      	 continue;
      }

      if ((entry = is_style_spec(spec,strlen(spec))) == (struct hashentry *) NULL) {
        STY_ERROR("Unknown style specifier: '%s'.",spec);
        skip_line(file);
        continue;
      }

      if ( !scan_args(file,entry->format,arg) ) {
	STY_ERROR("Too few arguments to style specifier: '%s'.",spec);
	skip_line(file);
	continue;
      }
        
      switch ( entry->symbol ) {

        case AOPEN:
	case ACLOSE:
	case ROPEN:
	case RCLOSE:
	case LEVEL:
	case QUOTE:
	case ACTUAL:
	case ENCAP:
	case ESCAPE:
		*((char *) entry->var1) = arg[0].c;
		break;

	case HEADINGS_FLAG:
		*((int *) entry->var1) = arg[0].d;
		break;

	case LINEMAX:
	case INDENT_LENGTH:
		if (arg[0].d >= 0)
		   *((int *) entry->var1) = arg[0].d;
		else
		   STY_ERROR2("%s must be positive (got %d)",spec,arg[0].d);
		break;

	case PREAMBLE:
	case POSTAMBLE:
	case GROUP_SKIP:
	case SETPAGEOPEN:
	case SETPAGECLOSE:
	case HEADING_PRE:
	case HEADING_SUF:
	case ITEM_0:
	case ITEM_1:
	case ITEM_2:
	case ITEM_01:
	case ITEM_x1:
	case ITEM_12:
	case ITEM_x2:
		*((char **) entry->var1) = xstrdup(arg[0].s);
		*((int *)   entry->var2)  = count_lfd(arg[0].s);
		break;

	case KEYWORD:
	case COMPOSITOR:
	case DELIM_0:
	case DELIM_1:
	case DELIM_2:
	case DELIM_N:
	case DELIM_R:
	case DELIM_T:
	case ENCAP_0:
	case ENCAP_1:
	case ENCAP_2:
	case INDENT_SPACE:
		*((char **) entry->var1) = xstrdup(arg[0].s);
		break;

	case PRECEDENCE:
		process_precedence(arg[0].s);
		break;

	case SORT_RULE:
		add_sort_rule(arg[0].s, arg[1].s, FALSE);
		break;

	case R_SORT_RULE:
		add_sort_rule(arg[0].s, arg[1].s, TRUE);
		break;

	case MERGE_RULE:
		add_merge_rule(arg[0].s, arg[1].s, FALSE);
		break;

	case R_MERGE_RULE:
		add_merge_rule(arg[0].s, arg[1].s, TRUE);
		break;

	case SORT_GROUP:
		add_group_rule(arg[0].s, arg[1].d);
		break;

	case GROUP_HEAD:
		add_heading(arg[0].d, arg[1].s);
		break;

	case INCLUDE_STY:
		scan_sty(arg[0].s);
		break;
      }
   }
}

static
int scan_args( file, format, arg )
FILE *file;
char *format;
arg_type *arg;
{
  int ok = TRUE;
  
  while (*format && ok) 
    if (*format++ == '%')
      switch (*format++) {
      	case 's':
		ok = scan_string( file, TRUE, arg->s );
		arg++;
		break;
	case 'r':
		ok = scan_string( file, FALSE, arg->s );
		arg++;
		break;
	case 'c':
		ok = scan_char(file, &arg->c );
		arg++;
		break;
	case 'd':
		ok = scan_int(file, &arg->d );
		arg++;
		break;
      }

   return ok;
}

static int
scan_spec(file, spec)
FILE *file;
char *spec;
{
  register char *pos = spec;

  do {
    *pos++ = TOLOWER(actual_char);
    next_char(file);
  } while (pos < &spec[STRING_MAX - 1] &&
	   actual_char != EOF && actual_char != SPC &&
	   actual_char != TAB && actual_char != LFD);

  *pos = NUL;

  if (pos == &spec[STRING_MAX - 1]) {
    STY_ERROR2("Specifier '%s' too long (max %d characters).\n",spec, STRING_MAX - 1);
    return (FALSE);
  }

  return (TRUE);
}

static
int scan_string( file, unquote, string )
FILE *file;
int unquote;
char *string;
{
   register char *s = string;
   
   skip_whitespace(file);

   if (actual_char != STR_DELIM) {
      STY_ERROR("Syntax error: String expected.","");
      return (FALSE);
   }

   while ( s < &string[LONG_MAX-1] &&
         next_char(file) != EOF && actual_char != STR_DELIM ) {
     if (actual_char == BSH)
	switch (next_char(file)) {
	  case 'n':	*s++ = LFD;   break;
	  case 't':	*s++ = TAB;   break;
	  case 'b':	*s++ = BEGIN; break;
	  case 'e':	*s++ = END;   break;
	  case 'x':	if (!scan_hexchar(file,s++)) return (FALSE);
			break;
	  default:
	  	if (!unquote) *s++ = BSH;
	  	*s++ = (char)actual_char;
	}
     else *s++ = (char)actual_char;
   }

   *s = NUL;
     
   if (s >= &string[LONG_MAX]) {
      STY_ERROR2("String '%s' too long (max %d characters).",string,LONG_MAX - 1);
      return (FALSE);
   }

   if (actual_char != STR_DELIM) {
      STY_ERROR("String delimiter expected.","");
      return (FALSE);
   }

   next_char(file);
   return (TRUE);
   
}

static 
int scan_char(file, ch_result)
FILE *file;
char *ch_result;
{
  /*
   * Assert that ch_result has a value.
   *	Note: I don't think this is needed. I did it while introducing
   * the usage of actual_char. The scanner must be rewritten anyhow,
   * I'll look at it later.
   */
  *ch_result = NUL;

  skip_whitespace(file);

  if (actual_char != CHR_DELIM) {
     STY_ERROR("Character argument expected.","");
     return (FALSE);
  }
  
  switch ( next_char(file) ) {
     case CHR_DELIM:
     case EOF:
       STY_ERROR("Character expected (premature closing delimiter or EOF)","");
       return (FALSE);
     case BSH:
       switch ( next_char(file) ) {
         case 'n':	actual_char = '\n'; break;
         case 't':	actual_char = '\t'; break;
	 case 'x':	if (!scan_hexchar(file,ch_result)) return (FALSE);
			break;
       }
  }

  *ch_result = (char)actual_char;

  if (next_char(file) != CHR_DELIM) {
    STY_ERROR("Character closing delimiter expected.","");
    return (FALSE);
  }

  next_char(file);
  return (TRUE);
}

static
int scan_hexchar(file, c)
FILE *file;
char *c;
{
  register int i;
  register int n;
  
  for (i = n = 0; i < 2; i++) {
    n *= 16;
    next_char(file);
    if (actual_char >= '0' && actual_char <= '9') n += actual_char - '0';
    else if (actual_char >= 'A' && actual_char <= 'F') n += actual_char + 10 - 'A';
    else if (actual_char >= 'a' && actual_char <= 'f') n += actual_char + 10 - 'a';
    else {
      STY_ERROR("Non hex-digit '%c' in escape sequence '\\x..'.",actual_char);
      return (FALSE);
    }
  }

  *c = (char) n;
  
  return (TRUE);
}

static
int scan_int(file, n)
FILE *file;
int *n;
{
  register int i = 0;
  register int sign = 1;

  skip_whitespace(file);

  if (actual_char == '+') next_char(file);
  else if (actual_char == '-') sign = -1, next_char(file);
  
  while (actual_char != EOF && actual_char >= '0' && actual_char <= '9') {
    i = i*10 + actual_char - '0';
    next_char(file);
  }

  *n = i * sign;

  return (actual_char == EOF ? FALSE : TRUE);  
}

static void
skip_whitespace(file)
FILE *file;
{
  while (TRUE)
    switch (next_char(file)) {
      case LFD:
      case SPC:
      case TAB:
	  break;
      case COMMENT:
          skip_line(file);
          break;
      default:
          return;
    }
}

static void
skip_line(file)
FILE *file;
{
  while (actual_char != LFD && actual_char != EOF) next_char(file);
}

static int
count_lfd(str)
char   *str;
{
  int n = 0;

  for (;*str;str++) if (*str == LFD) n++;
  return n;
}


static int
process_precedence( prec )
char *prec;
{
    int     order[PAGETYPE_MAX];
    int     type[PAGETYPE_MAX];
    int     i = 0;
    int     last;
    int     roml = FALSE;
    int     romu = FALSE;
    int     arab = FALSE;
    int     alpl = FALSE;
    int     alpu = FALSE;

    /* check for llegal specifiers first */
    while ((i < PAGETYPE_MAX) && (prec[i] != NUL)) {
        switch (prec[i]) {
        case ROMAN_LOWER:
            if (roml) {
                MULTIPLE(ROMAN_LOWER,prec);
            } else
                roml = TRUE;
            break;
        case ROMAN_UPPER:
            if (romu) {
                MULTIPLE(ROMAN_UPPER,prec);
            } else
                romu = TRUE;
            break;
        case ARABIC:
            if (arab) {
                MULTIPLE(ARABIC,prec);
            } else
                arab = TRUE;
            break;
        case ALPHA_LOWER:
            if (alpl) {
                MULTIPLE(ALPHA_UPPER,prec);
            } else
                alpl = TRUE;
            break;
        case ALPHA_UPPER:
            if (alpu) {
                MULTIPLE(ALPHA_UPPER,prec);
            } else
                alpu = TRUE;
            break;
        default:
            STY_ERROR("Unknow type `%c' in page precedence specification.\n",
                      prec[i]);
            return (FALSE);
        }
        i++;
    }
    if (prec[i] != NUL) {
        STY_ERROR("Page precedence specification string too long.\n", "");
        return (FALSE);
    }
    last = i;
    switch (prec[0]) {
    case ROMAN_LOWER:
        order[0] = ROMAN_LOWER_OFFSET;
        type[0] = ROML;
        break;
    case ROMAN_UPPER:
        order[0] = ROMAN_UPPER_OFFSET;
        type[0] = ROMU;
        break;
    case ARABIC:
        order[0] = ARABIC_OFFSET;
        type[0] = ARAB;
        break;
    case ALPHA_LOWER:
        order[0] = ALPHA_LOWER_OFFSET;
        type[0] = ALPL;
        break;
    case ALPHA_UPPER:
        order[0] = ALPHA_LOWER_OFFSET;
        type[0] = ALPU;
        break;
    }

    i = 1;
    while (i < last) {
        switch (prec[i]) {
        case ROMAN_LOWER:
            order[i] = order[i - 1] + ROMAN_LOWER_OFFSET;
            type[i] = ROML;
            break;
        case ROMAN_UPPER:
            order[i] = order[i - 1] + ROMAN_UPPER_OFFSET;
            type[i] = ROMU;
            break;
        case ARABIC:
            order[i] = order[i - 1] + ARABIC_OFFSET;
            type[i] = ARAB;
            break;
        case ALPHA_LOWER:
            order[i] = order[i - 1] + ALPHA_LOWER_OFFSET;
            type[i] = ALPL;
            break;
        case ALPHA_UPPER:
            order[i] = order[i - 1] + ALPHA_LOWER_OFFSET;
            type[i] = ALPU;
            break;
        }
        i++;
    }

    for (i = 0; i < PAGETYPE_MAX; i++) {
        page_offset[i] = -1;
    }
    page_offset[type[0]] = 0;
    for (i = 1; i < last; i++) {
        page_offset[type[i]] = order[i - 1];
    }
    for (i = 0; i < PAGETYPE_MAX; i++) {
        if (page_offset[i] == -1) {
            switch (type[last - 1]) {
            case ROML:
                order[last] = order[last - 1] + ROMAN_LOWER_OFFSET;
                break;
            case ROMU:
                order[last] = order[last - 1] + ROMAN_UPPER_OFFSET;
                break;
            case ARAB:
                order[last] = order[last - 1] + ARABIC_OFFSET;
                break;
            case ALPL:
                order[last] = order[last - 1] + ALPHA_LOWER_OFFSET;
                break;
            case ALPU:
                order[last] = order[last - 1] + ALPHA_UPPER_OFFSET;
                break;
            }
            type[last] = i;
            page_offset[i] = order[last];
            last++;
        }
    }
    return (TRUE);                     /* function value no longer used */
}



/*======================================================================
 *
 * $Log: scanst.c,v $
 * Revision 1.9  1992/01/03  13:22:34  schrod
 * next_char() returns int, not char; important for test on EOF. Discarded
 * intermediate variable, last read char is available in actual_char anyhow.
 *
 * Revision 1.8  1991/12/18  16:22:35  schrod
 * Enhanced handling of style files:
 *    Style file must always be there, a default (mkind.ist) is supplied.
 *    Default extension ".ist".
 *    Default search path to be configured in mkind.h. (Predefined is
 * ".:/usr/local/lib/makeindx", suitable on UNIX systems.)
 *
 * Now issue only short messages while reading style files. In a later
 * version a user should be able to turn long messages on again.
 *
 * Revision 1.7  1991/12/16  14:46:21  schrod
 * STYLE_PATH is accessed within set_stylepath() instead from the caller.
 * Implemented the search path access different, uses now less memory and
 * does not have an upper limit of elements in the path.
 *
 * Revision 1.6  1991/12/10  20:26:56  schrod
 * Completely rewritten the scanner to implement a C-like include feature.
 *
 * Revision 1.5  1991/12/10  20:10:06  schrod
 * Reject-rules (*merge_rule and *sort_rule) implemented.
 *
 * Revision 1.4  1991/12/10  19:20:43  schrod
 * Changed the group merging algorithm to create the group-id of an index key
 * using user defined rules (specified in the style file with commands:
 * `sort_group' and `group_head') instead of the sortkey's first character.
 *
 * Revision 1.3  1991/12/10  17:19:18  schrod
 * New version with flexible index sorting and merging. Style files now
 * can contain rules to describe, how to generate a sort or a merge key
 * from a given index entry.
 *
 * Revision 1.2  1991/12/05  20:08:40  schrod
 * MakeIndex, version 2.9.1.2 [by Joachim Schrod]
 * == base version, from which development of International MakeIndex starts ==
 *
 */

