/*
 * $Id: scanst.h,v 1.7 1991/12/16 14:46:26 schrod Exp $
 *----------------------------------------------------------------------
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 *
 * [history at end]
 */

#define COMMENT   '%'
#define STR_DELIM '"'
#define CHR_DELIM '\''

typedef enum {
   KEYWORD, AOPEN, ACLOSE, ROPEN, RCLOSE, LEVEL, QUOTE, ACTUAL, ENCAP, ESCAPE,
   COMPOSITOR, PRECEDENCE,
   PREAMBLE, POSTAMBLE, GROUP_SKIP, HEADINGS_FLAG, HEADING_PRE, HEADING_SUF,
   SETPAGEOPEN, SETPAGECLOSE, ITEM_0, ITEM_1, ITEM_2, ITEM_01, ITEM_x1,
   ITEM_12, ITEM_x2, DELIM_0, DELIM_1, DELIM_2, DELIM_N, DELIM_R, DELIM_T,
   ENCAP_0, ENCAP_1, ENCAP_2, LINEMAX, INDENT_SPACE, INDENT_LENGTH,
   SORT_RULE, MERGE_RULE, R_SORT_RULE, R_MERGE_RULE, SORT_GROUP, GROUP_HEAD,
   INCLUDE_STY
} symbol_type;

/* hashtable entry for the scanner keywords: */
struct hashentry {
   char *name; 			/* keyword name */
   symbol_type symbol; 		/* symbol number */
   char *format; 		/* format string, how to scan the parameters */
   VOIDP var1; 			/* optional pointer to variable, which is to set by this keyword */
   VOIDP var2;			/* like var1 */
};

/* the above format string consist of one or more sequences of the form
   %'char', one for each argument of the style specifier. 'char' describes
   the type of argument to scan:

   	's'	scan an in '"' enclosed string, and replace quoted characters
		'\.' by their values.
	'r'	like 's', but don't replace quoted characters; (useful for
		regular expressions).
	'c'	scan a character.
	'd'	scan an integer

   Other characters in the format string will be ignored!
*/

struct hashentry *is_style_spec ARGS((char *s, unsigned len));


/* Default style settings: */

#define SPEC_ARG_MAX	2

#define IDX_KEYWORD "\\indexentry"
#define IDX_AOPEN   '{'
#define IDX_ACLOSE  '}'
#define IDX_ROPEN   '('
#define IDX_RCLOSE  ')'
#define IDX_LEVEL   '!'
#define IDX_QUOTE   '"'
#define IDX_ACTUAL  '@'
#define IDX_ENCAP   '|'
#define IDX_ESCAPE  '\\'

#define COMPOSITOR_DEF     "-"
#define PRECEDENCE_DEF     "rnaRA"
#define ROMAN_LOWER        'r'
#define ROMAN_UPPER        'R'
#define ARABIC             'n'
#define ALPHA_LOWER        'a'
#define ALPHA_UPPER        'A'
#define ROMAN_LOWER_OFFSET 10000
#define ROMAN_UPPER_OFFSET 10000
#define ARABIC_OFFSET      10000
#define ALPHA_LOWER_OFFSET 26
#define ALPHA_UPPER_OFFSET 26

#define PREAMBLE_DEF  "\\begin{theindex}\n"
#define PREAMBLE_LEN  1
#define POSTAMBLE_DEF "\n\n\\end{theindex}\n"
#define POSTAMBLE_LEN 3

#if    KCC_20
/* KCC preprocessor bug collapses multiple blanks to single blank */
#define SETPAGEOPEN_DEF "\n\040\040\\setcounter{page}{"
#else
#define SETPAGEOPEN_DEF "\n  \\setcounter{page}{"
#endif                                 /* KCC_20 */

#define SETPAGECLOSE_DEF "}\n"
#define SETPAGE_LEN      2

#if    KCC_20
/* KCC preprocessor bug collapses multiple blanks to single blank */
#define GROUPSKIP_DEF "\n\n\040\040\\indexspace\n"
#else
#define GROUPSKIP_DEF "\n\n  \\indexspace\n"
#endif                                 /* KCC_20 */
#define GROUPSKIP_LEN 3

#define HEADINGSFLAG_DEF 0
#define HEADINGPRE_DEF   ""
#define HEADINGPRE_LEN   0
#define HEADINGSUF_DEF   ""
#define HEADINGSUF_LEN   0

#if    KCC_20
/* KCC preprocessor bug collapses multiple blanks to single blank */
#define ITEM0_DEF "\n\040\040\\item\040"
#define ITEM1_DEF "\n\040\040\040\040\\subitem\040"
#define ITEM2_DEF "\n\040\040\040\040\040\040\\subsubitem\040"
#else
#define ITEM0_DEF "\n  \\item "
#define ITEM1_DEF "\n    \\subitem "
#define ITEM2_DEF "\n      \\subsubitem "
#endif                                 /* KCC_20 */

#define ITEM_LEN 1

#define DELIM_DEF  ", "
#define DELIMR_DEF "--"
#define DELIMT_DEF ""

#define ENCAP0_DEF "\\"
#define ENCAP1_DEF "{"
#define ENCAP2_DEF "}"

#if    (OS_BS2000 | OS_MVSXA)
#define INDENTSPC_DEF "                "
#else
#define INDENTSPC_DEF "\t\t"
#endif                                 /* OS_BS2000 | OS_MVSXA */

#define INDENTLEN_DEF 16

#define STY_ERROR(F, D) { \
    if (idx_dot) { \
        fprintf(ilg_fp, "\n"); \
        idx_dot = FALSE; \
    } \
    fprintf(ilg_fp, "** Input style error (file = %s, line = %d):\n   -- ", \
            sty_fn, sty_lc); \
    fprintf(ilg_fp, F, D); \
    sty_ec++; \
    put_dot = FALSE; \
}

#if    KCC_20
/* KCC preprocessor bug collapses multiple blanks to single blank */
#define STY_ERROR2(F, D1, D2) { \
     if (idx_dot) { \
        fprintf(ilg_fp, "\n"); \
        idx_dot = FALSE; \
    } \
    fprintf(ilg_fp, \
            "** Input style error (file = %s, line = %d):\n\040\040 -- ", \
            sty_fn, sty_lc); \
    fprintf(ilg_fp, F, D1, D2); \
    sty_ec++; \
    put_dot = FALSE; \
}
#else
#define STY_ERROR2(F, D1, D2) { \
     if (idx_dot) { \
        fprintf(ilg_fp, "\n"); \
        idx_dot = FALSE; \
    } \
    fprintf(ilg_fp, "** Input style error (file = %s, line = %d):\n   -- ", \
            sty_fn, sty_lc); \
    fprintf(ilg_fp, F, D1, D2); \
    sty_ec++; \
    put_dot = FALSE; \
}
#endif                                 /* KCC_20 */

#define STY_DOT { \
    idx_dot = TRUE; \
    if (verbose) \
        fprintf(stderr, DOT); \
    fprintf(ilg_fp, DOT); \
}

#define MULTIPLE(C,P) { \
STY_ERROR2( \
"Multiple instances of type `%c' in page precedence specification `%s'.\n", \
C, P); \
    return (FALSE); \
}



/*======================================================================
 *
 * $Log: scanst.h,v $
 * Revision 1.7  1991/12/16  14:46:26  schrod
 * STYLE_PATH is accessed within set_stylepath() instead from the caller.
 * Implemented the search path access different, uses now less memory and
 * does not have an upper limit of elements in the path.
 *
 * Revision 1.6  1991/12/10  20:27:00  schrod
 * Completely rewritten the scanner to implement a C-like include feature.
 *
 * Revision 1.5  1991/12/10  20:10:10  schrod
 * Reject-rules (*merge_rule and *sort_rule) implemented.
 *
 * Revision 1.4  1991/12/10  19:20:48  schrod
 * Changed the group merging algorithm to create the group-id of an index key
 * using user defined rules (specified in the style file with commands:
 * `sort_group' and `group_head') instead of the sortkey's first character.
 *
 * Revision 1.3  1991/12/10  17:19:23  schrod
 * New version with flexible index sorting and merging. Style files now
 * can contain rules to describe, how to generate a sort or a merge key
 * from a given index entry.
 *
 * Revision 1.2  1991/12/05  20:08:45  schrod
 * MakeIndex, version 2.9.1.2 [by Joachim Schrod]
 * == base version, from which development of International MakeIndex starts ==
 *
 */

