/*
 * $Id: sortid.c,v 1.3 1991/12/10 17:19:27 schrod Exp $
 *----------------------------------------------------------------------
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 *
 * [history at end]
 */

#include    "mkind.h"

static int idx_gc;

static	int	compare ARGS((struct KFIELD * *a,struct KFIELD * *b));
static	int	compare_page ARGS((struct KFIELD * *a,struct KFIELD * *b));

void
sort_idx()
{
    MESSAGE("Sorting entries...", "");
    idx_dc = idx_gc = 0;
    qqsort((char *) idx_key, (int) idx_gt, (int) sizeof(FIELD_PTR), compare);
    MESSAGE("done (%d comparisons).\n", idx_gc);
}

static int
compare(a, b)
FIELD_PTR *a;
FIELD_PTR *b;
{
    int     i;
    int     dif;

    idx_gc++;
    IDX_DOT(CMP_MAX);

    for (i = 0; i < FIELD_MAX; i++) {
	/* compare the sort fields */
	if ((dif = strcmp((*a)->sf[i], (*b)->sf[i])) != 0)
	    break;
	/* compare the merge fields */
	if ((dif = strcmp((*a)->mf[i], (*b)->mf[i])) != 0)
	    break;
    }

    /* both key aggregates are identical, compare page numbers */
    if (i == FIELD_MAX)
	dif = compare_page(a, b);
    return (dif);
}

static int
compare_page(a, b)
FIELD_PTR *a;
FIELD_PTR *b;
{
    int     m;
    short   i = 0;

    while ((i < (*a)->count) && (i < (*b)->count) &&
	   ((m = (*a)->npg[i] - (*b)->npg[i]) == 0)) {
	i++;
    }
    if (m == 0) {
	if ((i == (*a)->count) && (i == (*b)->count)) {
	    if (STREQ((*a)->encap, (*b)->encap))
		/* identical entries */
		if (((*a)->type != DUPLICATE) &&
		    ((*b)->type != DUPLICATE))
		    (*b)->type = DUPLICATE;
		else if ((*(*a)->encap == idx_ropen) ||
			 (*(*b)->encap == idx_rclose))
		    m = -1;
		else if ((*(*a)->encap == idx_rclose) ||
			 (*(*b)->encap == idx_ropen))
		    m = 1;
	} else if ((i == (*a)->count) && (i < (*b)->count))
	    m = -1;
	else if ((i < (*a)->count) && (i == (*b)->count))
	    m = 1;
    }
    return (m);
}



/*======================================================================
 *
 * $Log: sortid.c,v $
 * Revision 1.3  1991/12/10  17:19:27  schrod
 * New version with flexible index sorting and merging. Style files now
 * can contain rules to describe, how to generate a sort or a merge key
 * from a given index entry.
 *
 * Revision 1.2  1991/12/05  20:08:50  schrod
 * MakeIndex, version 2.9.1.2 [by Joachim Schrod]
 * == base version, from which development of International MakeIndex starts ==
 *
 */

